/*	$NetBSD$	*/

/*-
 * Copyright (c) 2020 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <err.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <crypto/aes/aes.h>
#include <crypto/aes/aes_bear.h>
//#include <crypto/aes/aes_bear64.h>

#if defined(__i386__) || defined(__x86_64__)
#include <crypto/aes/arch/x86/aes_ni.h>
#include <crypto/aes/arch/x86/aes_sse2.h>
#include <crypto/aes/arch/x86/aes_ssse3.h>
#include <crypto/aes/arch/x86/aes_via.h>
#endif

#ifdef __aarch64__
#include <crypto/aes/arch/arm/aes_armv8.h>
#endif

#ifdef __ARM_NEON
#include <crypto/aes/arch/arm/aes_neon.h>
#endif

#define	arraycount(x)	(sizeof(x)/sizeof(x[0]))

enum cipher {
	CIPHER_NONE,
	CIPHER_AES_CBC,
	CIPHER_AES_XTS,
};

enum dir {
	DIR_NONE,
	DIR_ENCRYPT,
	DIR_DECRYPT,
};

static const struct aes_impl *const aes_impls[] = {
#ifdef __aarch64__
	&aes_armv8_impl,
#endif
//	&aes_bear64_impl,
	&aes_bear_impl,
#ifdef __ARM_NEON
	&aes_neon_impl,
#endif
#ifdef __x86_64__
	&aes_ni_impl,
#endif
#if defined(__i386__) || defined(__x86_64__)
	&aes_sse2_impl,
	&aes_ssse3_impl,
	&aes_via_impl,
#endif
};

static int
read_block(int fd, void *buf, size_t len)
{
	char *p = buf;
	size_t n = len;
	ssize_t nread;

	for (;;) {
		if ((nread = read(fd, p, n)) == -1)
			err(1, "read");
		if (nread == 0) {
			if (n < len)
				errx(1, "partial block");
			return -1; /* eof */
		}
		if ((size_t)nread >= n)
			break;
		p += (size_t)nread;
		n -= (size_t)nread;
	}

	return 0;
}

static void
write_block(int fd, const void *buf, size_t len)
{
	const char *p = buf;
	size_t n = len;
	ssize_t nwrit;

	for (;;) {
		if ((nwrit = write(fd, p, n)) == -1)
			err(1, "write");
		if ((size_t)nwrit >= n)
			break;
		p += (size_t)nwrit;
		n -= (size_t)nwrit;
	}
}

static void
go(enum cipher cipher, unsigned keybits, enum dir dir,
    const struct aes_impl *impl)
{
	static uint8_t key[64];
	static unsigned char buf[65536];
	struct aesenc tweakkey;
	struct aesenc *ivkey;
	struct aesenc enckey;
	struct aesdec deckey;
	void (*enc)(const struct aesenc *, const uint8_t[static 16],
	    uint8_t[static 16], size_t, uint8_t[static 16], uint32_t);
	void (*dec)(const struct aesdec *, const uint8_t[static 16],
	    uint8_t[static 16], size_t, uint8_t[static 16], uint32_t);
	uint8_t iv[16] __aligned(16) = {0};
	const unsigned blksize = 512;
	uint64_t blkno = 0;
	unsigned nrounds;

	switch (keybits) {
	case 128:
		nrounds = 10;
		break;
	case 192:
		nrounds = 12;
		break;
	case 256:
		nrounds = 14;
		break;
	default:
		abort();
	}
	(*impl->ai_setenckey)(&enckey, key, nrounds);
	(*impl->ai_setdeckey)(&deckey, key, nrounds);

	switch (cipher) {
	case CIPHER_AES_CBC:
		enc = impl->ai_cbc_enc;
		dec = impl->ai_cbc_dec;
		ivkey = &enckey;
		break;
	case CIPHER_AES_XTS:
		enc = impl->ai_xts_enc;
		dec = impl->ai_xts_dec;
		(*impl->ai_setenckey)(&tweakkey, key + keybits/8, nrounds);
		ivkey = &tweakkey;
		break;
	default:
		abort();
	}

	while (read_block(STDIN_FILENO, buf, sizeof(buf)) == 0) {
		unsigned i;

		for (i = 0; i < sizeof(buf); i += blksize, blkno++) {
			le64enc(iv, blkno);
			memset(iv + 8, 0, 8);
			(*impl->ai_enc)(ivkey, iv, iv, nrounds);

			switch (dir) {
			case DIR_ENCRYPT:
				(*enc)(&enckey, buf + i, buf + i, blksize, iv,
				    nrounds);
				break;
			case DIR_DECRYPT:
				(*dec)(&deckey, buf + i, buf + i, blksize, iv,
				    nrounds);
				break;
			default:
				abort();
			}
		}
		write_block(STDOUT_FILENO, buf, sizeof(buf));
	}
}

static void __dead
usage(void)
{

	fprintf(stderr,
	    "Usage: %s [-ed] -b <keybits> -c <cipher> -i <impl>\n",
	    getprogname());
	fprintf(stderr,
	    "       %s -l\n", getprogname());
	exit(1);
}

int
main(int argc, char **argv)
{
	const struct aes_impl *impl = NULL;
	enum cipher cipher = CIPHER_NONE;
	enum dir dir = DIR_NONE;
	unsigned keybits = 0;
	int ch, error = 0;

	while ((ch = getopt(argc, argv, "b:c:dei:l")) != -1) {
		switch (ch) {
		case 'b':
			if (strcmp(optarg, "128") == 0) {
				keybits = 128;
			} else if (strcmp(optarg, "192") == 0) {
				keybits = 192;
			} else if (strcmp(optarg, "256") == 0) {
				keybits = 256;
			} else {
				warnx("invalid key size in bits: %s", optarg);
				error = 1;
			}
			break;
		case 'c':
			if (strcmp(optarg, "aes-cbc") == 0) {
				cipher = CIPHER_AES_CBC;
			} else if (strcmp(optarg, "aes-xts") == 0) {
				cipher = CIPHER_AES_XTS;
			} else {
				warnx("invalid cipher: %s", optarg);
				error = 1;
			}
			break;
		case 'd':
			dir = DIR_DECRYPT;
			break;
		case 'e':
			dir = DIR_ENCRYPT;
			break;
		case 'i': {
			size_t i;

			for (i = 0; i < arraycount(aes_impls); i++) {
				if (strcmp(optarg, aes_impls[i]->ai_name) == 0)
					break;
			}
			if (i == arraycount(aes_impls)) {
				warnx("unknown aes implementation: %s",
				    optarg);
				error = 1;
				break;
			}
			impl = aes_impls[i];
			break;
		}
		case 'l': {
			size_t i;

			for (i = 0; i < arraycount(aes_impls); i++) {
				if ((*aes_impls[i]->ai_probe)())
					continue;
				if (aes_selftest(aes_impls[i]))
					continue;
				printf("%s\n", aes_impls[i]->ai_name);
			}
			fflush(stdout);
			exit(ferror(stdout));
		}
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc)
		usage();
	if (impl == NULL) {
		warnx("specify an AES implementation with `-i'");
		error = 1;
	}
	if (cipher == CIPHER_NONE) {
		warnx("specify a cipher with `-c'");
		error = 1;
	}
	if (dir == DIR_NONE) {
		warnx("specify a direction with `-e'/`-d'");
		error = 1;
	}
	if (keybits == 0) {
		warnx("specify a key size in bits with `-b'");
		error = 1;
	}
	if (error)
		usage();

	if ((*impl->ai_probe)())
		errx(1, "not supported on this hardware: %s", impl->ai_name);
	if (aes_selftest(impl))
		errx(1, "self-test failed: %s", impl->ai_name);

	go(cipher, keybits, dir, impl);
	return 0;
}
